import 'dart:async';

import 'package:meta/meta.dart';

import 'change.dart';
import 'core.dart';
import 'event_bus.dart';

abstract class Store<State> implements StoreBase<State> {
  final _completer = Completer<void>();
  final _disposables = <FutureOr<void> Function()>[];
  final List<StreamSubscription> _subscriptions = [];
  final StreamController<State> _stateController = StreamController<State>.broadcast();
  State _state;
  bool _emitted = false;

  Store(this._state);

  @override
  State get state => _state;

  @override
  bool get isClosed => _stateController.isClosed;

  @override
  Stream<State> get stream => _stateController.stream;

  @override
  @protected
  @mustCallSuper
  void emit(State state) {
    try {
      if (isClosed) {
        throw StateError('Cannot emit new states after calling close');
      }
      if (state == _state && _emitted) return;
      onChange(Change<State>(currentState: this.state, nextState: state));
      _state = state;
      _stateController.add(_state);
      _emitted = true;
    } catch (error, stackTrace) {
      onError(error, stackTrace);
      rethrow;
    }
  }

  @protected
  @mustCallSuper
  Future<void> emitOnEach<T>(
    Stream<T> stream, {
    required State Function(T data) onData,
    State Function(Object error, StackTrace stackTrace)? onError,
  }) {
    final completer = Completer<void>();
    var handleError = onError != null
        ? (Object error, StackTrace stackTrace) {
            emit(onError(error, stackTrace));
          }
        : null;
    final subscription = stream.listen(
      (data) => emit(onData(data)),
      onDone: completer.complete,
      onError: handleError ?? completer.completeError,
      cancelOnError: handleError == null,
    );
    _disposables.add(subscription.cancel);
    return Future.any([_completer.future, completer.future]).whenComplete(() {
      subscription.cancel();
      _disposables.remove(subscription.cancel);
    });
  }

  @protected
  @mustCallSuper
  pickup<E>(
    EventBus eventBus,
    EventListener<E> listener,
  ) {
    var subscription = eventBus.on(listener);
    _subscriptions.add(subscription);
  }

  @protected
  @mustCallSuper
  void onChange(Change<State> change) {}

  @protected
  @mustCallSuper
  void onError(Object error, StackTrace stackTrace) {}

  /// Closes the instance.
  /// This method should be called when the instance is no longer needed.
  /// Once [close] is called, the instance can no longer be used.
  @override
  @mustCallSuper
  Future<void> close() async {
    for (final disposable in _disposables) {
      disposable.call();
    }
    _disposables.clear();
    for (var subscription in _subscriptions) {
      subscription.cancel();
    }
    _subscriptions.clear();
    if (!_completer.isCompleted) _completer.complete();
    await _stateController.close();
  }
}
